#! /usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var yargs = _interopDefault(require('yargs'));
var vscodeLanguageserver = require('vscode-languageserver');
var fs = require('fs');
var nodeSqlParser = require('@joe-re/node-sql-parser');
var log4js = _interopDefault(require('log4js'));
var vscodeLanguageserverTypes = require('vscode-languageserver-types');
var vscodeJsonrpc = require('vscode-jsonrpc');
var EventEmitter = _interopDefault(require('events'));
var mysql = require('mysql');
var PG = _interopDefault(require('pg'));
var path = require('path');
var os = require('os');

class Cache {
    constructor() {
        this._cache = new Map();
    }
    get(uri) {
        let contents = this._cache.get(uri);
        return contents;
    }
    set(uri, contents) {
        this._cache.set(uri, contents);
    }
    setFromUri(uri) {
        this.set(uri, fs.readFileSync(uri, 'utf8'));
    }
}
var cache = new Cache();

const logger = log4js.getLogger();
const CLAUSES = [
    { label: 'WHERE', kind: vscodeLanguageserverTypes.CompletionItemKind.Text },
    { label: 'ORDER BY', kind: vscodeLanguageserverTypes.CompletionItemKind.Text },
    { label: 'GROUP BY', kind: vscodeLanguageserverTypes.CompletionItemKind.Text },
    { label: 'LIMIT', kind: vscodeLanguageserverTypes.CompletionItemKind.Text }
];
function extractExpectedLiterals(expected) {
    return expected.filter(v => v.type === 'literal')
        .map(v => v.text)
        .filter((v, i, self) => self.indexOf(v) === i)
        .map(v => ({ label: v }));
}
function getLastToken(sql) {
    const match = sql.match(/^(?:.|\s)*[^A-z0-9](.*?)$/);
    if (!match) {
        return sql;
    }
    return match[1];
}
function getColumnRefByPos(columns, pos) {
    return columns.find(v => (v.location.start.line === pos.line + 1 && v.location.start.column <= pos.column) &&
        (v.location.end.line === pos.line + 1 && v.location.end.column >= pos.column));
}
function getFromNodeByPos(fromNodes, pos) {
    return fromNodes.find(v => (v.location.start.line === pos.line + 1 && v.location.start.column <= pos.column) &&
        (v.location.end.line === pos.line + 1 && v.location.end.column >= pos.column));
}
function toCompletionItemFromTable(table) {
    return {
        label: table.tableName,
        detail: `table ${table.tableName}`,
        kind: vscodeLanguageserverTypes.CompletionItemKind.Text
    };
}
function toCompletionItemFromColumn(column) {
    return {
        label: column.columnName,
        detail: `column ${column.description}`,
        kind: vscodeLanguageserverTypes.CompletionItemKind.Text
    };
}
function getCandidatesFromColumnRefNode(columnRefNode, schema) {
    const tableCandidates = schema.filter(v => v.tableName.startsWith(columnRefNode.table)).map(v => toCompletionItemFromTable(v));
    const columnCandidates = Array.prototype.concat.apply([], schema.filter(v => tableCandidates.map(v => v.label).includes(v.tableName)).map(v => v.columns)).map((v) => toCompletionItemFromColumn(v));
    return tableCandidates.concat(columnCandidates);
}
function isCursorOnFromClause(sql, pos) {
    var _a;
    try {
        const ast = nodeSqlParser.parse(sql);
        return !!getFromNodeByPos(((_a = ast.from) === null || _a === void 0 ? void 0 : _a.tables) || [], pos);
    }
    catch (_e) {
        return false;
    }
}
function getCandidatedFromIncompleteSubquery(params) {
    let candidates = [];
    const { schema, incompleteSubquery, pos } = params;
    const parsedFromClause = getFromNodesFromClause(incompleteSubquery.text);
    try {
        nodeSqlParser.parse(incompleteSubquery.text);
    }
    catch (e) {
        if (e.name !== 'SyntaxError') {
            throw e;
        }
        const fromText = incompleteSubquery.text;
        const newPos = parsedFromClause ? {
            line: pos.line - (incompleteSubquery.location.start.line - 1),
            column: pos.column - incompleteSubquery.location.start.column + 1
        } : { line: 0, column: 0 };
        candidates = complete(fromText, newPos, schema).candidates;
    }
    return candidates;
}
function createTablesFromFromNodes(fromNodes) {
    return fromNodes.reduce((p, c) => {
        if (c.type !== 'subquery') {
            return p;
        }
        if (!Array.isArray(c.subquery.columns)) {
            return p;
        }
        const columns = c.subquery.columns.map(v => {
            if (typeof v === 'string') {
                return null;
            }
            return { columnName: v.as || v.expr.column || '', description: 'alias' };
        });
        return p.concat({ database: null, columns, tableName: c.as });
    }, []);
}
function getCandidatesFromError(target, schema, pos, e, fromNodes) {
    let candidates = extractExpectedLiterals(e.expected);
    const candidatesLiterals = candidates.map(v => v.label);
    if (candidatesLiterals.includes("'") || candidatesLiterals.includes('"')) {
        return [];
    }
    if (candidatesLiterals.includes('.')) {
        candidates = candidates.concat(schema.map(v => toCompletionItemFromTable(v)));
    }
    const lastChar = target[target.length - 1];
    logger.debug(`lastChar: ${lastChar}`);
    if (lastChar === '.') {
        const removedLastDotTarget = target.slice(0, target.length - 1);
        if (isCursorOnFromClause(removedLastDotTarget, { line: pos.line, column: pos.column - 1 })) {
            return [];
        }
        const tableName = getLastToken(removedLastDotTarget);
        const subqueryTables = createTablesFromFromNodes(fromNodes);
        const attachedAlias = schema.concat(subqueryTables).map(v => {
            const as = fromNodes.filter((v2) => v.tableName === v2.table).map(v => v.as);
            return Object.assign({}, v, { as: as ? as : [] });
        });
        let table = attachedAlias.find(v => v.tableName === tableName || v.as.includes(tableName));
        if (table) {
            candidates = table.columns.map(v => toCompletionItemFromColumn(v));
        }
    }
    return candidates;
}
function getFromNodesFromClause(sql) {
    try {
        return nodeSqlParser.parseFromClause(sql);
    }
    catch (_e) {
        // no-op
        return null;
    }
}
function getRidOfAfterCursorString(sql, pos) {
    return sql.split('\n').filter((_v, idx) => pos.line >= idx).map((v, idx) => idx === pos.line ? v.slice(0, pos.column) : v).join('\n');
}
function complete(sql, pos, schema = []) {
    var _a, _b, _c;
    logger.debug(`complete: ${sql}, ${JSON.stringify(pos)}`);
    let candidates = [];
    let error = null;
    const target = getRidOfAfterCursorString(sql, pos);
    logger.debug(`target: ${target}`);
    try {
        candidates = CLAUSES.concat([]);
        const ast = nodeSqlParser.parse(target);
        const ar = new nodeSqlParser.AstReader(ast);
        logger.debug(`ast: ${JSON.stringify(ar.getAst())}`);
        if (!ar.getAst().distinct) {
            candidates.push({ label: 'DISTINCT', kind: vscodeLanguageserverTypes.CompletionItemKind.Text });
        }
        const columns = ar.getAst().columns;
        if (Array.isArray(columns)) {
            const selectColumnRefs = columns.map((v) => v.expr).filter((v) => !!v);
            const whereColumnRefs = ar.getAst().where || [];
            const columnRef = getColumnRefByPos(selectColumnRefs.concat(whereColumnRefs), pos);
            logger.debug(JSON.stringify(columnRef));
            if (columnRef) {
                candidates = candidates.concat(getCandidatesFromColumnRefNode(columnRef, schema));
            }
        }
        if (Array.isArray((_a = ar.getAst().from) === null || _a === void 0 ? void 0 : _a.tables)) {
            const fromTable = getFromNodeByPos(((_b = ar.getAst().from) === null || _b === void 0 ? void 0 : _b.tables) || [], pos);
            if (fromTable && fromTable.type === 'table') {
                candidates = candidates.concat(schema.map(v => toCompletionItemFromTable(v)))
                    .concat([{ label: 'INNER JOIN' }, { label: 'LEFT JOIN' }]);
                if (fromTable.join && !fromTable.on) {
                    candidates.push({ label: 'ON' });
                }
            }
        }
    }
    catch (e) {
        logger.debug('error');
        logger.debug(e);
        if (e.name !== 'SyntaxError') {
            throw e;
        }
        const parsedFromClause = getFromNodesFromClause(sql);
        const fromNodes = ((_c = parsedFromClause === null || parsedFromClause === void 0 ? void 0 : parsedFromClause.from) === null || _c === void 0 ? void 0 : _c.tables) || [];
        const fromNodeOnCursor = getFromNodeByPos(fromNodes || [], pos);
        if (fromNodeOnCursor && fromNodeOnCursor.type === 'incomplete_subquery') {
            candidates = getCandidatedFromIncompleteSubquery({
                sql,
                pos,
                incompleteSubquery: fromNodeOnCursor,
                schema
            });
        }
        else {
            candidates = getCandidatesFromError(target, schema, pos, e, fromNodes);
        }
        error = { label: e.name, detail: e.message, line: e.line, offset: e.offset };
    }
    const lastToken = getLastToken(target);
    logger.debug(`lastToken: ${lastToken}`);
    logger.debug(JSON.stringify(candidates));
    candidates = candidates.filter(v => v.label.startsWith(lastToken));
    return { candidates, error };
}

const logger$1 = log4js.getLogger();
function createDiagnostics(uri, sql) {
    logger$1.debug(`createDiagnostics`);
    let diagnostics = [];
    try {
        const ast = nodeSqlParser.parse(sql);
        logger$1.debug(`ast: ${JSON.stringify(ast)}`);
    }
    catch (e) {
        logger$1.debug('parse error');
        logger$1.debug(e);
        if (e.name !== 'SyntaxError') {
            throw e;
        }
        diagnostics.push({
            range: {
                start: { line: e.location.start.line - 1, character: e.location.start.column },
                end: { line: e.location.end.line - 1, character: e.location.end.column }
            },
            message: e.message,
            severity: vscodeLanguageserverTypes.DiagnosticSeverity.Error,
            // code: number | string,
            source: 'sql',
            relatedInformation: []
        });
    }
    logger$1.debug(JSON.stringify(diagnostics));
    return { uri: uri, diagnostics };
}

const logger$2 = log4js.getLogger();
function createConnection(method) {
    logger$2.debug(`createConnection: method {${method}}`);
    switch (method) {
        case 'stdio': return vscodeLanguageserver.createConnection(process.stdin, process.stdout);
        case 'node-ipc':
        default:
            return vscodeLanguageserver.createConnection(new vscodeJsonrpc.IPCMessageReader(process), new vscodeJsonrpc.IPCMessageWriter(process));
    }
}

const logger$3 = log4js.getLogger();
class SettingStore extends EventEmitter {
    constructor() {
        super();
        this.state = {
            adapter: 'mysql',
            host: null,
            port: null,
            user: null,
            database: null,
            password: null
        };
    }
    static getInstance() {
        if (SettingStore.instance) {
            return SettingStore.instance;
        }
        SettingStore.instance = new SettingStore();
        return SettingStore.instance;
    }
    getSetting() {
        return Object.assign({}, this.state);
    }
    setSettingFromFile(path) {
        return new Promise((resolve, reject) => {
            fs.stat(path, (err, _stat) => {
                if (err) {
                    if (err.code && err.code === 'ENOENT') {
                        logger$3.debug("there isn't config file.");
                        resolve();
                        return;
                    }
                    logger$3.error(err.message);
                    reject(err);
                    return;
                }
                fs.readFile(path, { encoding: 'utf8' }, (err, data) => {
                    if (err) {
                        logger$3.error(err.message);
                        reject(err);
                        return;
                    }
                    this.setSetting(JSON.parse(data));
                    logger$3.debug(`read settings from ${path}`);
                    logger$3.debug(JSON.stringify(this.getSetting()));
                    resolve(this.getSetting());
                    return;
                });
            });
        });
    }
    setSetting(setting) {
        this.state = Object.assign({}, this.state, setting);
        this.emit('change', this.state);
    }
}

const logger$4 = log4js.getLogger();
class AbstractClient {
    constructor(settings) {
        this.settings = settings;
    }
    async getSchema() {
        let schema = [];
        try {
            this.connect();
            const tables = await this.getTables();
            schema = await Promise.all(tables.map((v) => this.getColumns(v).then(columns => ({
                database: this.settings.database,
                tableName: v,
                columns: columns.map(v => this.toColumnFromRawField(v))
            }))));
            this.disconnect();
        }
        catch (e) {
            this.disconnect();
            logger$4.error(e);
            throw e;
        }
        return schema;
    }
    toColumnFromRawField(field) {
        return {
            columnName: field.field,
            description: `${field.field}(Type: ${field.type}, Null: ${field.null}, Default: ${field.default})`
        };
    }
}

class MysqlClient extends AbstractClient {
    constructor(settings) {
        super(settings);
        this.connection = null;
    }
    connect() {
        this.connection = mysql.createConnection({
            host: this.settings.host || 'localhost',
            password: this.settings.password || '',
            user: this.settings.user || 'root',
            port: this.settings.port || 3306,
            database: this.settings.database || undefined
        });
    }
    disconnect() {
        if (this.connection) {
            this.connection.end();
        }
        this.connection = null;
    }
    getTables() {
        const sql = `
      SELECT table_name 
      FROM information_schema.tables
      WHERE table_schema = '${this.settings.database}'
    `;
        return new Promise((resolve, reject) => {
            if (!this.connection) {
                reject(new Error("Don't have database connection."));
                return;
            }
            this.connection.query(sql, (err, results) => {
                if (err) {
                    reject(new Error(err.message));
                    return;
                }
                const tables = results.map((v) => v[`table_name`]);
                resolve(tables);
            });
        });
    }
    getColumns(tableName) {
        const sql = `SHOW FULL FIELDS FROM ${tableName}`;
        return new Promise((resolve, reject) => {
            if (!this.connection) {
                reject(new Error("Don't have database connection."));
                return;
            }
            this.connection.query(sql, (err, results) => {
                if (err) {
                    reject(new Error(err.message));
                    return;
                }
                const columns = results.map((v) => ({
                    field: v.Field,
                    type: v.Type,
                    null: v.Null,
                    default: v.Default,
                    comment: v.Comment
                }));
                resolve(columns);
            });
        });
    }
}

class PosgresClient extends AbstractClient {
    constructor(settings) {
        super(settings);
        this.connection = null;
    }
    connect() {
        const client = new PG.Client({
            user: this.settings.user || '',
            host: this.settings.host || '',
            database: this.settings.database || '',
            password: this.settings.password || '',
            port: this.settings.port || 5432,
        });
        client.connect();
        this.connection = client;
    }
    disconnect() {
        if (this.connection) {
            this.connection.end();
        }
        this.connection = null;
    }
    getTables() {
        const sql = `
      SELECT c.relname as table_name FROM pg_class c LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
       WHERE n.nspname = 'public'
         AND c.relkind IN ('r','v','m','f')
    `;
        return new Promise((resolve, reject) => {
            if (!this.connection) {
                reject(new Error("Don't have database connection."));
                return;
            }
            this.connection.query(sql, (err, results) => {
                if (err) {
                    reject(new Error(err.message));
                    return;
                }
                const tables = results.rows.map((v) => v[`table_name`]);
                resolve(tables);
            });
        });
    }
    getColumns(tableName) {
        const sql = `
    SELECT
      a.attname as field,
      format_type(a.atttypid, a.atttypmod) as type,
      pg_get_expr(d.adbin, d.adrelid) as default,
      a.attnotnull as null,
      col_description(a.attrelid, a.attnum) AS comment
    FROM pg_attribute a
      LEFT JOIN pg_attrdef d ON a.attrelid = d.adrelid AND a.attnum = d.adnum
      LEFT JOIN pg_type t ON a.atttypid = t.oid
      LEFT JOIN pg_collation c ON a.attcollation = c.oid AND a.attcollation <> t.typcollation
    WHERE a.attrelid = '${tableName}'::regclass
      AND a.attnum > 0 AND NOT a.attisdropped
    ORDER BY a.attnum
    `;
        return new Promise((resolve, reject) => {
            if (!this.connection) {
                reject(new Error("Don't have database connection."));
                return;
            }
            this.connection.query(sql, (err, results) => {
                if (err) {
                    reject(new Error(err.message));
                    return;
                }
                resolve(results.rows);
            });
        });
    }
}

function getDatabaseClient(settings) {
    switch (settings.adapter) {
        case 'mysql': return new MysqlClient(settings);
        case 'postgresql': return new PosgresClient(settings);
        default: throw new Error(`not support ${settings.adapter}`);
    }
}

const MAX_LOG_SIZE = 1024 * 1024;
const MAX_LOG_BACKUPS = 10;
const LOG_FILE_PATH = path.join(os.tmpdir(), 'sql-language-server.log');
function initializeLogging() {
    log4js.configure({
        appenders: {
            server: {
                type: 'file',
                filename: LOG_FILE_PATH,
                axLogSize: MAX_LOG_SIZE,
                ackups: MAX_LOG_BACKUPS
            }
        },
        categories: { default: { appenders: ['server'], level: yargs.argv.debug ? 'debug' : 'warn' } }
    });
    const logger = log4js.getLogger();
    process.on('uncaughtException', e => logger.error('uncaughtException', e));
    process.on('unhandledRejection', e => logger.error('unhandledRejection', e));
    return logger;
}

function createServer() {
    let connection = createConnection(yargs.argv.method || 'node-ipc');
    initializeLogging();
    const logger = log4js.getLogger();
    let documents = new vscodeLanguageserver.TextDocuments();
    documents.listen(connection);
    let schema = [];
    connection.onInitialize((params) => {
        logger.debug(`onInitialize: ${params.rootPath}`);
        if (params.rootPath) {
            SettingStore.getInstance().setSettingFromFile(`${params.rootPath}/.sqllsrc.json`);
        }
        SettingStore.getInstance().on('change', async () => {
            try {
                const client = getDatabaseClient(SettingStore.getInstance().getSetting());
                schema = await client.getSchema();
                logger.debug('get schema');
                logger.debug(JSON.stringify(schema));
            }
            catch (e) {
                logger.error(e);
            }
        });
        return {
            capabilities: {
                textDocumentSync: documents.syncKind,
                completionProvider: {
                    resolveProvider: true,
                    triggerCharacters: ['.'],
                }
            }
        };
    });
    connection.onDidChangeTextDocument((params) => {
        logger.debug(`didChangeTextDocument: ${params.textDocument.uri}`);
        cache.set(params.textDocument.uri, params.contentChanges[0].text);
        const diagnostics = createDiagnostics(params.textDocument.uri, params.contentChanges[0].text);
        connection.sendDiagnostics(diagnostics);
    });
    connection.onCompletion((docParams) => {
        let text = cache.get(docParams.textDocument.uri);
        if (!text) {
            cache.setFromUri(docParams.textDocument.uri);
            text = cache.get(docParams.textDocument.uri);
        }
        logger.debug(text || '');
        const candidates = complete(text || '', {
            line: docParams.position.line,
            column: docParams.position.character
        }, schema).candidates;
        logger.debug(candidates.map(v => v.label).join(","));
        return candidates;
    });
    connection.onCompletionResolve((item) => {
        return item;
    });
    connection.listen();
    logger.info('start server');
    return connection;
}

const cli = yargs
    .usage('SQL Language Server Command Line Interface')
    .command('up', 'run sql-language-server', {
    method: {
        alias: 'm',
        type: 'string',
        default: 'node-ipc',
        choices: ['stdio', 'node-ipc'],
        describe: 'What use to communicate with sql language server'
    },
    'debug': {
        alias: 'd',
        type: 'boolean',
        default: false,
        describe: 'Enable debug logging'
    }
}, () => {
    createServer();
    process.stdin.resume();
})
    .example('$0 up --method stdio', ': start up sql-language-server - communicate via stdio')
    .help()
    .argv;
if (cli._.length === 0) {
    yargs.showHelp();
}
process.stdin.on('close', () => {
    process.exit(0);
});
//# sourceMappingURL=cli.js.map
